/*
 * Copyright 1996-2000 Hans Reiser
 */
#define __USE_GNU
#include <stdio.h>

#include "fsck.h"
#include <getopt.h>
#include <sys/mount.h>



#define print_usage_and_exit() die ("Usage: %s [-aprvy] [--rebuild-tree] device\n\
\n\
-v verbose mode\n\
-a supress progress information\n\
--rebuild-tree force fsck to rebuild filesystem from scratch\n\
        (takes a long time)
-y\n\
-p do nothing, exist for compatibility with fsck(8)\n\
-r\n", argv[0]);

int opt_verbose = 0;
int opt_check = 1; /* default behavior */
int opt_rebuild = 0;
int opt_fsck = 0; /* called with -a by fsck - the front-end for the
		     various file system checkers */

int opt_find = 0;
struct key key_to_find;


#define WARNING \
"Don't run this program unless something is broken.  You may want\n\
to backup first.  Some types of random FS damage can be recovered\n\
from by this program, which basically throws away the internal nodes\n\
of the tree and then reconstructs them.  This program is for use only\n\
by the desperate, and is of only beta quality.  Email\n\
reiserfs@devlinux.com with bug reports.  Do you want to run this\n\
program?[N/Yes] (note need to type Yes):"

/* 
   warning #2
   you seem to be running this automatically.  you are almost
   certainly doing it by mistake as a result of some script that
   doesn't know what it does.  doing nothing, rerun without -p if you
   really intend to do this.  */


/* fsck is called with one non-optional argument - file name of device
   containing reiserfs. This function parses other options, sets flags
   based on parsing and returns non-optional argument */
static char * parse_options (int argc, char * argv [])
{
    int c;
    char * answer = 0;
    size_t n = 0;

    while (1) {
	static struct option options[] = {
	    {"rebuild-tree", no_argument, &opt_rebuild, 1},
	    {0, 0, 0, 0}
	};
	int option_index;
      
	c = getopt_long (argc, argv, "yapv", options, &option_index);
	if (c == -1)
	    break;

	switch (c) {
	case 0:
	    if (option_index == 0) {
		printf (WARNING);
		if (getline (&answer, &n, stdin) != 4 || strcmp ("Yes\n", answer)) {
		    exit (0);
		}
		opt_rebuild = 1;
		opt_check = 0;
	    }
	    break;

	case 'y':
	case 'p': /* these do nothing */
	case 'r':
	    break;

	case 'a':
	    opt_fsck = 1;
	    break;

	case 'v':
	    /* output fsck statistics to stdout on exit */
	    opt_verbose = 1;
	    break;
#if 0
	case 'i':
	    opt_find = 1;
	    /* get key to search */
	    printf ("Input key to search (set offset and uniqueness to -1 to look for any items of file:\n");
	    if (scanf ("%d %d %d %d", &(key_to_find.k_dir_id), &(key_to_find.k_objectid),
		       &(key_to_find.k_offset), &(key_to_find.k_uniqueness)) != 4)
		die ("parse_options: specify a key through stdin");
	    break;
#endif

	default:
	    print_usage_and_exit();
	}
    }
  if (optind != argc - 1)
    /* only one non-option argument is permitted */
    print_usage_and_exit();
  
  return argv[optind];
}


struct super_block g_sb;
struct buffer_head * g_sbh;
struct reiserfs_super_block * g_old_rs;



static void reset_super_block (struct super_block * s)
{
  unsigned long * oids;

  g_old_rs = (struct reiserfs_super_block *)kmalloc (s->s_blocksize);
  memcpy (g_old_rs, SB_BUFFER_WITH_SB (s)->b_data, s->s_blocksize);

  /* reset few fields in */
  SB_FREE_BLOCKS (s) = SB_BLOCK_COUNT (s);
  SB_TREE_HEIGHT (s) = ~0;
  SB_ROOT_BLOCK (s) = ~0;
  s->u.reiserfs_sb.s_mount_state = REISERFS_ERROR_FS;
  s->u.reiserfs_sb.s_rs->s_oid_cursize = 2;
  oids = (unsigned long *)(s->u.reiserfs_sb.s_rs + 1);
  if (oids[0] != 1) {
    printf ("reiserfsck_read_super_block: invalid objectid map\n");
    oids[0] = 1;
  }
  oids[1] = oids[0];
  s->s_dirt = 1;

  mark_buffer_dirty (SB_BUFFER_WITH_SB (s), 0);
}


static void release_super_block ()
{
  SB_REISERFS_STATE (&g_sb) = REISERFS_VALID_FS;


  mark_buffer_dirty (SB_BUFFER_WITH_SB (&g_sb), 0);
  freemem (SB_AP_BITMAP (&g_sb));
  brelse (SB_BUFFER_WITH_SB (&g_sb));

  freemem (g_old_rs);
}


char ** g_disk_bitmap;
char ** g_new_bitmap;
char ** g_uninsertable_leaf_bitmap;
char ** g_formatted;
char ** g_unformatted;
int g_blocks_to_read;


/* read bitmaps (new or old format), create data blocks for new
   bitmap, mark non-data blocks in it (skipped, super block, journal
   area, bitmaps) used, create other auxiliary bitmaps */
static void init_bitmaps (int old_format)
{
    int i, j;

    /* read disk bitmap */
/*
    if (old_format ? read_old_bitmaps (&g_sb) : read_bitmaps(&g_sb))
	die ("init_bitmap: unable to read bitmap");
*/
    if (read_bitmaps(&g_sb))
	die ("init_bitmap: unable to read bitmap");

    g_disk_bitmap = kmalloc (sizeof (char *) * SB_BMAP_NR (&g_sb));
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++) {
	g_disk_bitmap[i] = SB_AP_BITMAP (&g_sb)[i]->b_data;
    }

    /* calculate number of blocks marked in bitmap as used */
    g_blocks_to_read = 0;
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++) {
	for (j = 0; j < g_sb.s_blocksize * 8; j ++)
	    if (i * g_sb.s_blocksize * 8 + j < SB_BLOCK_COUNT (&g_sb) &&
		test_bit (j, SB_AP_BITMAP (&g_sb)[i]->b_data))
		g_blocks_to_read ++;
    }

    /* this bitmap will contain valid bitmap when fsck will have done */
    g_new_bitmap = kmalloc (sizeof (char *) * SB_BMAP_NR (&g_sb));
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++)
	g_new_bitmap[i] = kmalloc (g_sb.s_blocksize);

    /* mark skipped blocks, super block, journal area and first bitmap
       as used */
/*    for (i = 0; i < SB_JOURNAL_BLOCK(&g_sb) + JOURNAL_BLOCK_COUNT + 1; i ++)*/
    for (i = 0; i < g_sb.u.reiserfs_sb.s_rs->s_journal_block + JOURNAL_BLOCK_COUNT + 1; i ++)
	mark_block_used (i);
    /* mark other bitmap blocks as used */
    for (i = 1; i < SB_BMAP_NR (&g_sb); i ++)
	mark_block_used (SB_AP_BITMAP (&g_sb)[i]->b_blocknr);
    /* fill by 1s the unused part of last bitmap */
    if (SB_BLOCK_COUNT (&g_sb) % (g_sb.s_blocksize * 8))
	for (j = SB_BLOCK_COUNT (&g_sb) % (g_sb.s_blocksize * 8); j < g_sb.s_blocksize * 8; j ++)
	    set_bit (j, g_new_bitmap[SB_BMAP_NR (&g_sb) - 1]);

    /* allocate space for bitmap of uninsertable leaves */
    g_uninsertable_leaf_bitmap = kmalloc (sizeof (char *) * SB_BMAP_NR (&g_sb));
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++) {
	g_uninsertable_leaf_bitmap[i] = kmalloc (g_sb.s_blocksize);
	memset (g_uninsertable_leaf_bitmap[i], 0xff, g_sb.s_blocksize);
    }

    /* bitmap of formatted nodes */
    g_formatted = kmalloc (sizeof (char *) * SB_BMAP_NR (&g_sb));
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++) {
	g_formatted[i] = kmalloc (g_sb.s_blocksize);
	memset (g_formatted[i], 0, g_sb.s_blocksize);
    }
    /* bitmap of unformatted nodes */
    g_unformatted = kmalloc (sizeof (char *) * SB_BMAP_NR (&g_sb));
    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++) {
	g_unformatted[i] = reiserfs_kmalloc (g_sb.s_blocksize);
	memset (g_unformatted[i], 0, g_sb.s_blocksize);
    }
}


/* write bitmaps and brelse them */
static void update_bitmap ()
{
    int i;


    for (i = 0; i < SB_BMAP_NR (&g_sb); i ++) {

	/* copy newly built bitmap to cautious bitmap */
	memcpy (SB_AP_BITMAP (&g_sb)[i]->b_data, g_new_bitmap[i], g_sb.s_blocksize);
	mark_buffer_dirty (SB_AP_BITMAP (&g_sb)[i], 0);
	bwrite (SB_AP_BITMAP (&g_sb)[i]);
    
	/* release bitmap buffers */
	brelse (SB_AP_BITMAP (&g_sb)[i]);

	freemem (g_new_bitmap[i]);
	/* g_disk_bitmap[i] points to corresponding cautious bitmap's b_data */
	freemem (g_uninsertable_leaf_bitmap[i]);
    }

    freemem (g_disk_bitmap);
    freemem (g_new_bitmap);
    freemem (g_uninsertable_leaf_bitmap);

}


static void replay_journal (char * devname1)
{
    int retval;
    char * tmpdir;

    printf ("Replaying journal.."); fflush (stdout);

    tmpdir = tmpnam (0);
    if (!tmpdir || mkdir (tmpdir, 0644))
	die ("replay_journal: tmpnam or mkdir failed: %s", strerror (errno));
 
    retval = mount (devname1, tmpdir, "reiserfs", MS_MGC_VAL, "replayonly");
    if (retval != -1 || errno != EINVAL) {
	printf ("\nMAKE SURE, THAT YOUR KERNEL IS ABLE TO MOUNT REISERFS\n");
	die ("replay_journal: mount returned unexpected value: %s", 
	     strerror (errno));
    }

    if (rmdir (tmpdir) == -1)
	die ("replay_journal: rmdir failed: %s", strerror (errno));

    printf ("ok\n"); fflush (stdout);
}


static inline int nothing_todo (struct super_block * s)
{
    if (!opt_check && !opt_rebuild)
	return 1;
    return 0;
}


void write_dirty_blocks (void)
{
  fsync_dev (0);
}


int main (int argc, char * argv [])
{
  char * file_name;
  int old_format = 0;

  if (opt_fsck == 0)
    printf ("\n\n<-----------REISERFSCKv2, 1999----------->\n\n");


  file_name = parse_options (argc, argv);
  if (is_mounted (file_name))
      /* do not work on mounted filesystem for now */
      die ("reiserfsck: '%s' contains a mounted file system\n", file_name);

/*  replay_journal (file_name);*/

  g_sb.s_dev = open (file_name, O_RDWR);
  if (g_sb.s_dev == -1)
      die ("reiserfsck: can not open '%s': %s", file_name, strerror (errno));
  
  if (read_super_block (&g_sb, 1024))
      die ("reiserfsck: no reiserfs (v2) found after journal replay success");

  if (nothing_todo (&g_sb)) {
      printf ("%s: clean, %d/%d %ldk blocks\n", file_name,
	      SB_BLOCK_COUNT (&g_sb) - SB_FREE_BLOCKS(&g_sb), SB_BLOCK_COUNT (&g_sb), g_sb.s_blocksize / 1024);
      brelse (SB_BUFFER_WITH_SB (&g_sb));
      return 0;
  }
      
  if (opt_check) {
      check_fs_tree (&g_sb, old_format);
      return 0;
  }

  if (opt_fsck == 1)
    printf ("ReiserFS : checking %s\n",file_name);
  else
    printf ("Rebuilding..\n");

  /* re-building */
  reset_super_block (&g_sb);
  init_bitmaps (old_format);

  /* make file system invalid unless fsck done */
  SB_REISERFS_STATE (&g_sb) = REISERFS_ERROR_FS;
  bwrite (SB_BUFFER_WITH_SB (&g_sb));

  /* 1,2. building of the tree */
  build_the_tree ();

  /* 3. semantic pass */
  semantic_pass ();

  /* 4. look for unaccessed items in the leaves */
  check_unaccessed_items ();

  update_bitmap ();
  release_super_block ();

  if (opt_verbose == 1)
    output_information ();

  printf ("Syncing.."); fflush (stdout);

  write_dirty_blocks ();
  sync ();

  printf ("done\n"); fflush (stdout);
  
  if (opt_verbose == 1)
    printf ("Checking mem..");

  free_overwritten_unfms ();
/*  check_and_free_buffer_mem ();*/

  if (opt_verbose == 1)
    printf ("done\n");

  if (opt_fsck == 1)
    printf("ReiserFS : done checking %s\n",file_name);
  else
    printf ("Ok\n");

  return 0;
}
