/*
 * Copyright 1996, 1997, 1998 Hans Reiser, see reiserfs/README for licensing and copyright details
 */
#ifdef __KERNEL__

#include <linux/sched.h>
#include <linux/reiserfs_fs.h>
#include <linux/smp_lock.h>

#else

#include "nokernel.h"

#endif



#define MAX_GEN_NUMBER  127


#define SET_GENERATION_NUMBER(offset,gen_number) (GET_HASH_VALUE(offset)|(gen_number))


/* key is pointer to cpu key */
static int bin_search_in_dir_item (struct item_head * ih, struct reiserfs_de_head * deh, 
				   struct cpu_key * key, int * pos_in_item)
{
    int rbound, lbound, j;

    lbound = 0;
    rbound = I_ENTRY_COUNT (ih) - 1;

    for (j = (rbound + lbound) / 2; lbound <= rbound; j = (rbound + lbound) / 2) {
	if (cpu_key_k_offset (key) < le32_to_cpu (deh[j].deh_offset)) {
	    rbound = j - 1;
	    continue;
	}
	if (cpu_key_k_offset (key) > le32_to_cpu (deh[j].deh_offset)) {
	    lbound = j + 1;
	    continue;
	}
	/* key found */
	*pos_in_item = j;
	return POSITION_FOUND;
    }

    *pos_in_item = lbound;
    return POSITION_NOT_FOUND;
}


/* first calls search_by_key, then, if item is not found looks for the
   entry inside directory item found by search_by_key. (We assign a
   key to each directory item, and place multiple entries in a single
   directory item.)  Fills the path to the entry, and to the entry
   position in the item */
/* The function is NOT SCHEDULE-SAFE! */
int search_by_entry_key (struct super_block * sb, struct cpu_key * key, struct path * path)
{
    struct buffer_head * bh;
    struct item_head * ih;

    /* search for a directory item using key of entry */
    if (search_item (sb, key, path) == ITEM_FOUND) {
	path->pos_in_item = 0;
	return POSITION_FOUND;
    }
#ifdef REISERFS_CHECK
    if (!PATH_LAST_POSITION (path))
	reiserfs_panic (sb, "vs-7000: search_by_entry_key: search_by_key returned item position == 0");
#endif /* REISERFS_CHECK */
    PATH_LAST_POSITION (path) --;

    bh = get_bh (path);
    ih = get_ih (path);

#ifdef REISERFS_CHECK
    if (!is_direntry_le_ih (ih) || COMP_SHORT_KEYS (&(ih->ih_key), key)) {
	print_block (bh, 0, -1, -1);
	reiserfs_panic (sb, "vs-7005: search_by_entry_key: found item %h is not directory item or "
			"does not belong to the same directory as key %k", ih, key);
    }
#endif /* REISERFS_CHECK */

    /* binary search in directory item by third component of the key */
    return bin_search_in_dir_item (ih, B_I_DEH (bh, ih), key, &(path->pos_in_item));
}



/* Keyed 32-bit hash function using TEA in a Davis-Meyer function */
static unsigned long get_third_component (const char * name, int len)
{
    unsigned long res;

    if (!len || (len == 1 && name[0] == '.'))
	return DOT_OFFSET;
    if (len == 2 && name[0] == '.' && name[1] == '.')
	return DOT_DOT_OFFSET;

    res = keyed_hash (name, len);
    res = GET_HASH_VALUE(res);
    if (res == 0)
	res = 128;
    return res + MAX_GEN_NUMBER;
}


/* fills the structure with various parameters of directory entry,
   including key of the pointed object */
static void get_entry_attributes (struct reiserfs_dir_entry * de, int entry_num)
{
#ifdef REISERFS_CHECK
    if (I_ENTRY_COUNT (de->de_ih) < entry_num)
	reiserfs_panic (0, "yr-7006: get_entry_attributes: no such entry (%d-th) in the item (%d)",
			entry_num, I_ENTRY_COUNT (de->de_ih));
    if (de->de_deh != B_I_DEH (de->de_bh, de->de_ih) + entry_num)
	reiserfs_panic (0, "yr-7008: get_entry_attributes: dir entry header not found");
    
#endif /* REISERFS_CHECK */

    /* few fields are set already (de_bh, de_item_num, de_deh) */
    de->de_entrylen = entry_length (de->de_bh, de->de_ih, entry_num);
    de->de_namelen = de->de_entrylen - (de_with_sd (de->de_deh) ? SD_SIZE : 0);

    de->de_name = B_I_PITEM (de->de_bh, de->de_ih) + le16_to_cpu (de->de_deh->deh_location);
    if (de->de_name[de->de_namelen - 1] == 0) {
	/* name is padded with 0s to the 4 byte boundary */
	de->de_namelen = strlen (de->de_name);
    }

    /* key of object pointed by entry */
    de->de_dir_id = le32_to_cpu (de->de_deh->deh_dir_id);
    de->de_objectid = le32_to_cpu (de->de_deh->deh_objectid);

    /* key of the entry */
    de->de_entry_key.version = ITEM_VERSION_1;
    de->de_entry_key.on_disk_key.k_dir_id = le32_to_cpu (de->de_ih->ih_key.k_dir_id);
    de->de_entry_key.on_disk_key.k_objectid = le32_to_cpu (de->de_ih->ih_key.k_objectid);
    set_cpu_key_k_offset (&(de->de_entry_key), le32_to_cpu (de->de_deh->deh_offset));
    set_cpu_key_k_type (&(de->de_entry_key), TYPE_DIRENTRY);
}


static int try_name (struct reiserfs_dir_entry * de, 
		     const char * name,
		     int          namelen)
{
    int retval = POSITION_NOT_FOUND;

    if ((namelen == de->de_namelen) &&
	!memcmp(de->de_name, name, de->de_namelen))
	retval = de_visible (de->de_deh) ? POSITION_FOUND : POSITION_FOUND_INVISIBLE;

    return retval;
}


/* after this function de_entry_num is set correctly only if name
   found or there was no entries with given hash value */
static int linear_search_in_dir_item (struct cpu_key * key, struct reiserfs_dir_entry * de,
				      const char * name, int namelen)
{
    int retval;
    int i;

    i = de->de_entry_num;

    if (i == I_ENTRY_COUNT (de->de_ih) ||
	GET_HASH_VALUE (le32_to_cpu (de->de_deh[i].deh_offset)) != GET_HASH_VALUE (cpu_key_k_offset (key))) {
	i --;
    }

#ifdef REISERFS_CHECK
    if (de->de_deh != B_I_DEH (de->de_bh, de->de_ih))
	reiserfs_panic (0, "vs-7010: linear_search_in_dir_item: array of entry headers not found");
#endif /* REISERFS_CHECK */

    de->de_deh += i;

    for (; i >= 0; i --, de->de_deh --) {
	if (GET_HASH_VALUE (le32_to_cpu (de->de_deh->deh_offset)) !=
	    GET_HASH_VALUE (cpu_key_k_offset (key))) {
	    return POSITION_NOT_FOUND;
	}
   
	/* mark, that this generation number is used */
	if (de->de_gen_number_bit_string)
	    set_bit (GET_GENERATION_NUMBER (de->de_deh->deh_offset), de->de_gen_number_bit_string);

	/* de_bh, de_item_num, de_ih, de_deh are already set. Set others fields */
	get_entry_attributes (de, i);
	if ((retval = try_name (de, name, namelen)) != POSITION_NOT_FOUND) {
	    de->de_entry_num = i;
	    return retval;
	}
    }

    if (GET_GENERATION_NUMBER (le_key_k_offset (ITEM_VERSION_1, &(de->de_ih->ih_key))) == 0)
	/* we have reached left most entry in the node. In common we
           have to go to the left neighbor, but if generation counter
           is 0 already, we know for sure, that there is no name with
           the same hash value */
	return POSITION_NOT_FOUND;

#ifdef REISERFS_CHECK
    if (le_key_k_offset (ITEM_VERSION_1, &(de->de_ih->ih_key)) <= DOT_DOT_OFFSET
	|| de->de_item_num != 0)
	/* first entry can not be neither "." nor ".." and item we
           looked through must be 0-th item in the node */
	reiserfs_panic (0, "vs-7015: linear_search_in_dir_item: "
			"directory %k corrupted", key);
#endif /* REISERFS_CHECK */

    return GOTO_PREVIOUS_ITEM;
}


static int reiserfs_find_entry (struct inode * dir, const char * name, int namelen, 
				struct path * path_to_entry, struct reiserfs_dir_entry * de)
{
    struct cpu_key key_to_search;
    int retval;

    if (!dir || !dir->i_sb)
	return POSITION_NOT_FOUND;

    if ((unsigned int)namelen > REISERFS_MAX_NAME_LEN (dir->i_sb->s_blocksize))
	return POSITION_NOT_FOUND;

    /* we will search for this key in the tree */
    make_cpu_key (&key_to_search, dir, 
		  get_third_component (name, namelen), TYPE_DIRENTRY, 3);

    while (1) {
	/* search for a directory item using the formed key */
	if (search_item (dir->i_sb, &key_to_search, path_to_entry) == ITEM_NOT_FOUND) {
	    /* take previous item */
#ifdef REISERFS_CHECK
	    if (!PATH_LAST_POSITION (path_to_entry))
		reiserfs_panic (dir->i_sb, "vs-7010: reiserfs_find_entry: search_by_key returned bad position == 0");
#endif /* REISERFS_CHECK */
	    PATH_LAST_POSITION (path_to_entry) --;
	}
    
	de->de_bh = PATH_PLAST_BUFFER (path_to_entry);
	de->de_item_num = PATH_LAST_POSITION (path_to_entry);
	de->de_ih = B_N_PITEM_HEAD (de->de_bh, de->de_item_num);
	de->de_deh = B_I_DEH (de->de_bh, de->de_ih);

#ifdef REISERFS_CHECK
	if (!is_direntry_le_ih (de->de_ih) || comp_short_le_keys (&(de->de_ih->ih_key), INODE_PKEY (dir)))
	    reiserfs_panic (dir->i_sb, "vs-7020: reiserfs_find_entry: "
			    "item must be an item of the same directory item as inode");
#endif /* REISERFS_CHECK */

	/* we do not check whether bin_search_in_dir_item found the given
	   key, even if so, we still have to compare names */
	bin_search_in_dir_item (de->de_ih, de->de_deh, &key_to_search, &(de->de_entry_num));

	/* compare names for all entries having given hash value */
	retval = linear_search_in_dir_item (&key_to_search, de, name, namelen);
	if (retval != GOTO_PREVIOUS_ITEM) {
	    /* there is no need to scan directory anymore. Given entry found or does not exist */
	    path_to_entry->pos_in_item = de->de_entry_num;
	    return retval;
	}

	/* there is left neighboring item of this directory and given entry can be there */
	set_cpu_key_k_offset (&key_to_search, le_key_k_offset (ITEM_VERSION_1, &(de->de_ih->ih_key)) - 1);
	pathrelse (path_to_entry);

    } /* while (1) */
}



/* add entry to the directory (entry can be hidden). Does not mark dir
   inode dirty, do it after successesfull call to it */
static int reiserfs_add_entry (struct reiserfs_transaction_handle *th, struct inode * dir, 
                               const char * name, int namelen, struct key * object_key,
			       struct reiserfs_dir_entry * de,
			       int visible)
{
    struct cpu_key entry_key;
    char * buffer;
    int buflen;
    struct reiserfs_de_head * deh;
    INITIALIZE_PATH (path);
    char bit_string [MAX_GEN_NUMBER / 8 + 1];
    int gen_number;


    if (!dir || !dir->i_sb)
	return -ENOENT;

    if ((unsigned int)namelen > REISERFS_MAX_NAME_LEN (dir->i_sb->s_blocksize))
	return -ENAMETOOLONG;

    /* each entry has unique key. compose it */
    make_cpu_key (&entry_key, dir, 
		  get_third_component (name, namelen), TYPE_DIRENTRY, 3);

    /* get memory for composing the entry */
    buflen = DEH_SIZE + ROUND_UP (namelen);
    buffer = reiserfs_kmalloc (buflen, GFP_KERNEL, dir->i_sb);
    if (buffer == 0)
	return -ENOMEM;

    /* fill buffer : directory entry head, name[, dir objectid | , stat data | ,stat data, dir objectid ] */
    deh = (struct reiserfs_de_head *)buffer;
    deh->deh_location = 0;
    deh->deh_offset = cpu_to_le32 (cpu_key_k_offset (&entry_key));
    deh->deh_state = 0;
    /* put key (ino analog) to de */
    deh->deh_dir_id = object_key->k_dir_id;
    deh->deh_objectid = object_key->k_objectid;

    /* copy name */
    memcpy ((char *)(deh + 1), name, namelen);
    /* padd by 0s to the 4 byte boundary */
    padd_item ((char *)(deh + 1), ROUND_UP (namelen), namelen);

    /* entry is ready to be pasted into tree, set 'visibility' and 'stat data in entry' attributes */
    mark_de_without_sd (deh);
    visible ? mark_de_visible (deh) : mark_de_hidden (deh);

    /* find the proper place for the new entry */
    memset (bit_string, 0, sizeof (bit_string));
    de->de_gen_number_bit_string = bit_string;
    if (reiserfs_find_entry (dir, name, namelen, &path, de) == POSITION_FOUND) {
	reiserfs_panic (dir->i_sb, "vs-7030: reiserfs_add_entry: entry with this key %k already exists",
			&entry_key);
    }

    if (find_first_nonzero_bit (bit_string, MAX_GEN_NUMBER + 1) < MAX_GEN_NUMBER + 1) {
	/* there are few names with given hash value */
	gen_number = find_first_zero_bit (bit_string, MAX_GEN_NUMBER + 1);
	if (gen_number > MAX_GEN_NUMBER) {
	    /* there is no free generation number */
	    reiserfs_kfree (buffer, buflen, dir->i_sb);
	    pathrelse (&path);
	    return -EEXIST;
	}
	/* adjust offset of directory enrty */
	deh->deh_offset = cpu_to_le32 (SET_GENERATION_NUMBER (le32_to_cpu (deh->deh_offset), gen_number));
	set_cpu_key_k_offset (&entry_key, deh->deh_offset);

	/* find place for new entry */
	if (search_by_entry_key (dir->i_sb, &entry_key, &path)) {
	    reiserfs_panic (dir->i_sb, "reiserfs_add_entry: 7032: entry with this key (%k) already exists", &entry_key);
	}
	de->de_entry_num = path.pos_in_item;
    } else {
	deh->deh_offset = cpu_to_le32 (SET_GENERATION_NUMBER (le32_to_cpu (deh->deh_offset), 0));
	set_cpu_key_k_offset (&entry_key, le32_to_cpu (deh->deh_offset));
    }
  
    /* perform the insertion of the entry that we have prepared */
    if (reiserfs_paste_into_item (th, dir->i_sb, &path, &entry_key, buffer, buflen) == -1) {
	reiserfs_kfree (buffer, buflen, dir->i_sb);
	reiserfs_check_path(&path) ;
	return -ENOSPC;
    }

    reiserfs_kfree (buffer, buflen, dir->i_sb);
    dir->i_size += buflen;
    dir->i_blocks = ((dir->i_size + 511) >> 9);
    dir->i_mtime = dir->i_ctime = CURRENT_TIME;
    reiserfs_check_path(&path) ;
    return 0;
}


struct dentry * reiserfs_lookup (struct inode * dir, struct dentry * dentry)
{
    struct inode * inode = 0;
    struct reiserfs_dir_entry de;
    INITIALIZE_PATH (path_to_entry);
    int error;

    reiserfs_check_lock_depth("lookup") ;

    de.de_gen_number_bit_string = 0;
    error = reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &de);
    pathrelse (&path_to_entry);
    if (error == POSITION_FOUND) {
	inode = reiserfs_iget (dir->i_sb, (struct cpu_key *)&(de.de_dir_id));
	if (!inode) {
	    return ERR_PTR(-EACCES);
        }
    }

    d_add(dentry, inode);
    return NULL;
}


int reiserfs_create (struct inode * dir, struct dentry *dentry, int mode)
{
    int error;
    struct inode * inode;
    struct reiserfs_dir_entry de;
    int err;
    int windex ;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 2 ;
    struct reiserfs_transaction_handle th ;
	
    if (!dir)
	return -ENOENT;
	
    inode = get_empty_inode() ;
    if (!inode) {
	return -ENOSPC ;
    }
    journal_begin(&th, dir->i_sb, jbegin_count) ;
    th.t_caller = "create" ;
    windex = push_journal_writer("reiserfs_create") ;
    inode = reiserfs_new_inode (&th, dir, mode, 0, 0, dentry, inode, &err);
    if (!inode) {
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	return err;
    }
	
    inode->i_op = &reiserfs_file_inode_operations;
    inode->i_fop = &reiserfs_file_operations;
    inode->i_mapping->a_ops = &reiserfs_address_space_operations ;

    error = reiserfs_add_entry (&th, dir, dentry->d_name.name, dentry->d_name.len, INODE_PKEY (inode), &de, 1);
    if (error) {
	inode->i_nlink--;
	reiserfs_update_sd (&th, inode, READ_BLOCKS);
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	iput (inode);
	return error;
    }
    // stat data of a directory file is created in gets into
    // transaction as well
    reiserfs_update_sd (&th, dir, READ_BLOCKS);
    d_instantiate(dentry, inode);
    pop_journal_writer(windex) ;
    journal_end(&th, dir->i_sb, jbegin_count) ;
    return 0;
}


int reiserfs_mknod (struct inode * dir, struct dentry *dentry, int mode, int rdev)
{
    int error;
    struct inode * inode;
    INITIALIZE_PATH (path_to_entry);
    struct reiserfs_dir_entry de;
    int err;
    int windex ;
    struct reiserfs_transaction_handle th ;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3; 

    if (!dir)
	return -ENOENT;

    inode = get_empty_inode() ;
    if (!inode) {
      return -ENOSPC ;
    }
    journal_begin(&th, dir->i_sb, jbegin_count) ;
    windex = push_journal_writer("reiserfs_mknod") ;
    
    de.de_gen_number_bit_string = 0;
    if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &de) == POSITION_FOUND) {
	pathrelse ( &path_to_entry);
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	iput(inode) ;
	return -EEXIST;
    }

    pathrelse ( &path_to_entry);

    inode = reiserfs_new_inode (&th, dir, mode, 0, 0, dentry, inode, &err);
    if (!inode) {
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	reiserfs_check_path(&path_to_entry) ;
	return err;
    }

    inode->i_uid = current->fsuid;
    inode->i_mode = mode;
    inode->i_op = NULL;

    if (S_ISREG(inode->i_mode))
	inode->i_op = &reiserfs_file_inode_operations;
    else if (S_ISDIR(inode->i_mode)) {
	inode->i_op = &reiserfs_dir_inode_operations;
	if (dir->i_mode & S_ISGID)
	    inode->i_mode |= S_ISGID;
    }
    else if (S_ISLNK(inode->i_mode))
	inode->i_op = &reiserfs_symlink_inode_operations;
    else {
      init_special_inode(inode, inode->i_mode, rdev) ;
    }

    reiserfs_update_sd (&th, inode, READ_BLOCKS);

    error = reiserfs_add_entry (&th, dir, dentry->d_name.name, dentry->d_name.len, INODE_PKEY (inode), &de, 1);
    if (error) {
	inode->i_nlink--;
	reiserfs_update_sd (&th, inode, READ_BLOCKS);
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	iput (inode);
	reiserfs_check_path(&path_to_entry) ;
	return error;
    }
    reiserfs_update_sd (&th, dir, READ_BLOCKS);
    d_instantiate(dentry, inode);
    pop_journal_writer(windex) ;
    journal_end(&th, dir->i_sb, jbegin_count) ;
    reiserfs_check_path(&path_to_entry) ;
    return 0;
}


int reiserfs_mkdir (struct inode * dir, struct dentry *dentry, int mode)
{
    int error;
    struct inode * inode;
    INITIALIZE_PATH (path_to_entry);
    struct reiserfs_dir_entry de;
    int windex ;
    struct reiserfs_transaction_handle th ;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3; 
    int err;

    if (!dir || !dir->i_sb) {
	return -EINVAL;
    }

    inode = get_empty_inode() ;
    if (!inode) {
      return -ENOSPC ;
    }
    journal_begin(&th, dir->i_sb, jbegin_count) ;
    windex = push_journal_writer("reiserfs_mkdir") ;
    de.de_gen_number_bit_string = 0;
    if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &de) == POSITION_FOUND) {
	pathrelse (&path_to_entry);
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	iput(inode) ;
	return -EEXIST;
    }
    pathrelse (&path_to_entry);
  
    if (dir->i_nlink >= REISERFS_LINK_MAX) {
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	iput(inode) ;
	reiserfs_check_path(&path_to_entry) ;
	return -EMLINK;
    }
  
    mode = S_IFDIR | (mode & 0777 & ~current->fs->umask);
    if (dir->i_mode & S_ISGID)
	mode |= S_ISGID;
    inode = reiserfs_new_inode (&th, dir, mode, 0, 0, dentry, inode, &err);
    if (!inode) {
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	reiserfs_check_path(&path_to_entry) ;
	return err;
    }

    inode->i_op = &reiserfs_dir_inode_operations;
    inode->i_fop = &reiserfs_dir_operations;

    /* new inode and stat data are uptodate. Inode is clean. */
    error = reiserfs_add_entry (&th, dir, dentry->d_name.name, dentry->d_name.len, INODE_PKEY (inode), &de, 1);
    if (error) {
	inode->i_nlink = 0;
	reiserfs_update_sd (&th, inode, READ_BLOCKS);
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	iput (inode);
	reiserfs_check_path(&path_to_entry) ;
	return error;
    }

    /* update dir inode, reiserfs_add_entry does not do that */
    dir->i_nlink ++;
    reiserfs_update_sd (&th, dir, READ_BLOCKS);
    d_instantiate(dentry, inode);
    pop_journal_writer(windex) ;
    journal_end(&th, dir->i_sb, jbegin_count) ;
    reiserfs_check_path(&path_to_entry) ;
    return 0;
}


static int reiserfs_empty_dir(struct inode * inode)
{
    return inode->i_size == EMPTY_DIR_SIZE;
}


int rmdir_not_allowed (struct inode * dir, struct reiserfs_dir_entry * de, struct inode * inode)
{
    if ((dir->i_mode & S_ISVTX) && !fsuser() &&
	current->fsuid != inode->i_uid &&
	current->fsuid != dir->i_uid)
	return -EPERM;

    if (inode->i_dev != dir->i_dev)
	return -EPERM;

    if (inode == dir)	/* we may not delete ".", but "../dir" is ok */
	return -EPERM;

    if (!S_ISDIR(inode->i_mode))
	return -ENOTDIR;

    if (!reiserfs_empty_dir(inode))
	return -ENOTEMPTY;

    if (de->de_objectid != inode->i_ino)
	return -ENOENT;

    return 0;
}


int reiserfs_rmdir (struct inode * dir, struct dentry *dentry)
{
    struct inode * inode;
    int retval;
    struct reiserfs_dir_entry de;
    int windex ;
    struct reiserfs_transaction_handle th ;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3; 
    INITIALIZE_PATH (path);


    retval = -ENOENT;
    de.de_gen_number_bit_string = 0;
    journal_begin(&th, dir->i_sb, jbegin_count) ;
    windex = push_journal_writer("reiesrfs_rmdir") ;
    if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path, &de) == POSITION_NOT_FOUND)
	goto end_rmdir;

    inode = dentry->d_inode;

    retval = rmdir_not_allowed (dir, &de, inode);
    if (retval)
	goto end_rmdir;

    if (!reiserfs_empty_dir (inode))
	retval = -ENOTEMPTY;
    else {
	/* cut entry from dir directory */
	if (reiserfs_cut_from_item (&th, dir, &path, &(de.de_entry_key), 0) == -1) {
	    retval = -ENOENT;
	}
    }
    if (retval)
	goto end_rmdir;

    if (inode->i_nlink != 2)
	printk ("reiserfs_rmdir: empty directory has nlink != 2 (%d)\n", inode->i_nlink);
    inode->i_nlink = 0;
    inode->i_ctime = dir->i_ctime = dir->i_mtime = CURRENT_TIME;
    reiserfs_update_sd (&th, inode, READ_BLOCKS);
    dir->i_nlink --;
    dir->i_size -= (DEH_SIZE + de.de_entrylen);
    dir->i_blocks = ((dir->i_size + 511) >> 9);
    reiserfs_update_sd (&th, dir, READ_BLOCKS);

    pop_journal_writer(windex) ;
    journal_end(&th, dir->i_sb, jbegin_count) ;
    d_delete(dentry); /* note, must be after journal end */
    reiserfs_check_path(&path) ;
    return 0;
	
 end_rmdir:
    /* we must release path, because we did not call
       reiserfs_cut_from_item, or reiserfs_cut_from_item does not
       release path if operation was not complete */
    pathrelse (&path);
    pop_journal_writer(windex) ;
    journal_end(&th, dir->i_sb, jbegin_count) ;
    return retval;	
}


int reiserfs_unlink (struct inode * dir, struct dentry *dentry)
{
    int retval;
    struct inode * inode;
    struct reiserfs_dir_entry de;
    INITIALIZE_PATH (path);
    int windex ;
    int call_journal_end = 1 ;
    struct reiserfs_transaction_handle th ;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3; 

    retval = -ENOENT;

    journal_begin(&th, dir->i_sb, jbegin_count) ;
    windex = push_journal_writer("reiserfs_unlink") ;
	
    de.de_gen_number_bit_string = 0;
    if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path, &de) == POSITION_NOT_FOUND) {
	goto end_unlink;
    }
    inode = dentry->d_inode;

    retval = -EPERM;
    if (S_ISDIR (inode->i_mode)) {
	goto end_unlink;
    }
    if ((dir->i_mode & S_ISVTX) && !fsuser() &&
	current->fsuid != inode->i_uid &&
	current->fsuid != dir->i_uid) {
	goto end_unlink;
    }

    retval = -ENOENT;
    if (comp_short_keys (INODE_PKEY (inode), (struct cpu_key *)&(de.de_dir_id))) {
	goto end_unlink;
    }
  
    if (!inode->i_nlink) {
	printk("reiserfs_unlink: deleting nonexistent file (%s:%lu), %d\n",
	       kdevname(inode->i_dev), inode->i_ino, inode->i_nlink);
	inode->i_nlink = 1;
    }

    if (reiserfs_cut_from_item (&th, dir, &path, &(de.de_entry_key), 0) == -1) {
	retval = -ENOENT;
	goto end_unlink;
    }

    inode->i_nlink--;
    inode->i_ctime = CURRENT_TIME;
    reiserfs_update_sd (&th, inode, READ_BLOCKS);
    dir->i_size -= (de.de_entrylen + DEH_SIZE);
    dir->i_blocks = ((dir->i_size + 511) >> 9);
    dir->i_ctime = dir->i_mtime = CURRENT_TIME;
    reiserfs_update_sd (&th, dir, READ_BLOCKS);
    pathrelse (&path);
    pop_journal_writer(windex) ;

    journal_end(&th, dir->i_sb, jbegin_count) ;
    call_journal_end = 0 ;
    d_delete(dentry);
    retval = 0;

 end_unlink:
    if (call_journal_end) {
	pathrelse (&path);
	pop_journal_writer(windex) ;
        journal_end(&th, dir->i_sb, jbegin_count) ;
    }
    reiserfs_check_path(&path) ;
    return retval;
}


int reiserfs_symlink (struct inode * dir, struct dentry * dentry, const char * symname)
{
    struct inode * inode;
    INITIALIZE_PATH (path_to_entry);
    struct reiserfs_dir_entry de;
    int err;
    char * name;
    int item_len;
    int windex ;
    struct reiserfs_transaction_handle th ;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3; 

    inode = get_empty_inode() ;
    if (!inode) {
      return -ENOSPC ;
    }

    item_len = ROUND_UP (strlen (symname));
    if (item_len > MAX_ITEM_LEN (dir->i_sb->s_blocksize)) {
	return -ENAMETOOLONG;
    }
  
    name = kmalloc (item_len, GFP_KERNEL);
    if (!name)
	return -ENOMEM;
    memcpy (name, symname, strlen (symname));
    padd_item (name, item_len, strlen (symname));

    journal_begin(&th, dir->i_sb, jbegin_count) ;
    windex = push_journal_writer("reiserfs_symlink") ;

    inode = reiserfs_new_inode (&th, dir, S_IFLNK, name, item_len, dentry, inode, &err);
    kfree (name);
    if (inode == 0) {
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	reiserfs_check_path(&path_to_entry) ;
	return err;
    }
    inode->i_op = &reiserfs_symlink_inode_operations;
    inode->i_size = strlen (symname);
    inode->i_mode = S_IFLNK | 0777;
    reiserfs_update_sd (&th, inode, READ_BLOCKS);

    de.de_gen_number_bit_string = 0;
    if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &de) == POSITION_FOUND) {
	pathrelse (&path_to_entry);
	inode->i_nlink--;
	reiserfs_update_sd (&th, inode, READ_BLOCKS);
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	iput (inode);
	return -EEXIST;
    }
    pathrelse (&path_to_entry);

    err = reiserfs_add_entry (&th, dir, dentry->d_name.name, dentry->d_name.len, INODE_PKEY (inode), &de, 1);
    if (err) {
	inode->i_nlink--;
	reiserfs_update_sd (&th, inode, READ_BLOCKS);
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	iput (inode);
	reiserfs_check_path(&path_to_entry) ;
	return err;
    }
    reiserfs_update_sd (&th, dir, READ_BLOCKS);
    d_instantiate(dentry, inode);
    pop_journal_writer(windex) ;
    journal_end(&th, dir->i_sb, jbegin_count) ;
    reiserfs_check_path(&path_to_entry) ;
    return 0;
}


int reiserfs_link (struct dentry * old_dentry, struct inode * dir, struct dentry * dentry)
{
    struct inode *inode = old_dentry->d_inode;
    INITIALIZE_PATH (path_to_entry);
    struct reiserfs_dir_entry de;
    int error;
    int windex ;
    struct reiserfs_transaction_handle th ;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3; 
  
    /* object must not be directory */
    if (S_ISDIR(inode->i_mode)) {
	return -EPERM;
    }
  
    /* file has too many links */
    if (inode->i_nlink >= REISERFS_LINK_MAX) {
	return -EMLINK;
    }

    journal_begin(&th, dir->i_sb, jbegin_count) ;
    windex = push_journal_writer("reiserfs_link") ;
    de.de_gen_number_bit_string = 0;

    if (reiserfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len, &path_to_entry, &de) == POSITION_FOUND) {
	pathrelse (&path_to_entry);
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	return -EEXIST;
    }
  
    pathrelse (&path_to_entry);

    /* create new entry */
    error = reiserfs_add_entry (&th, dir, dentry->d_name.name, dentry->d_name.len, INODE_PKEY (inode), &de, 1);
    if (error) {
	pop_journal_writer(windex) ;
	journal_end(&th, dir->i_sb, jbegin_count) ;
	reiserfs_check_path(&path_to_entry) ;
	return error;
    }
    inode->i_nlink++;
    inode->i_ctime = CURRENT_TIME;

    reiserfs_update_sd (&th, inode, READ_BLOCKS);
    reiserfs_update_sd (&th, dir, READ_BLOCKS);

    inode->i_count++;
    d_instantiate(dentry, inode);
    pop_journal_writer(windex) ;
    journal_end(&th, dir->i_sb, jbegin_count) ;
    reiserfs_check_path(&path_to_entry) ;
    return 0;
}


static int de_still_valid (const char * name, int len, struct reiserfs_dir_entry * de)
{
    struct item_head * ih;
    struct reiserfs_de_head * deh;

    if (!de || !de->de_bh)
	return 0;

    deh = B_I_DEH (de->de_bh, ih = B_N_PITEM_HEAD (de->de_bh, de->de_item_num));
    /* compare dir entry headers, record loacation and names */
    if (memcmp (&(deh[de->de_entry_num]), de->de_deh, DEH_SIZE) ||
	B_I_E_NAME (de->de_bh, ih, de->de_entry_num) != de->de_name ||
	memcmp (name, de->de_name, len))
	return 0;
    return 1;
}


static int entry_points_to_object (const char * name, int len, struct reiserfs_dir_entry * de, struct inode * inode)
{
    if (!de_still_valid (name, len, de))
	return 0;

    if (inode) {
	if (!de_visible (de->de_deh))
	    reiserfs_panic (0, "vs-7042: entry_points_to_object: entry must be visible");
	return (de->de_objectid == inode->i_ino) ? 1 : 0;
    }

    /* this must be added hidden entry */
    if (de_visible (de->de_deh))
	reiserfs_panic (0, "vs-7043: entry_points_to_object: entry must be visible");

    return 1;
}


/* sets key of parent directory in ".." entry */
static void set_ino_in_dir_entry (struct reiserfs_dir_entry * de, struct key * key)
{
    de->de_deh->deh_dir_id = key->k_dir_id;
    de->de_deh->deh_objectid = key->k_objectid;
}

/* 
 * process, that is going to call fix_nodes/do_balance must hold only
 * one path. If it holds 2 or more, it can get into endless waiting in
 * get_empty_nodes or its clones 
 */
int reiserfs_rename (struct inode * old_dir, struct dentry *old_dentry,
		     struct inode * new_dir, struct dentry *new_dentry)
{
    int retval;
    INITIALIZE_PATH (old_entry_path);
    INITIALIZE_PATH (new_entry_path);
    INITIALIZE_PATH (dot_dot_entry_path);
    struct reiserfs_dir_entry old_de, new_de, dot_dot_de;
    struct inode * old_inode, * new_inode;
    int new_entry_added = 0;
    int windex ;
    struct reiserfs_transaction_handle th ;
    int jbegin_count = JOURNAL_PER_BALANCE_CNT * 3; 

    journal_begin(&th, old_dir->i_sb, jbegin_count) ;
    windex = push_journal_writer("reiserfs_rename") ;

    goto start_up;

 try_again:
    run_task_queue(&tq_disk);
    current->policy |= SCHED_YIELD;
    /*current->counter = 0;*/
    schedule();
	
 start_up:
    old_inode = new_inode = NULL;
    dot_dot_de.de_bh = 0;
    new_de.de_bh = 0;

    /* 
     * look for the old name in old directory 
     */
    retval = -ENOENT;
    old_de.de_gen_number_bit_string = 0;
    if (reiserfs_find_entry (old_dir, old_dentry->d_name.name, old_dentry->d_name.len, &old_entry_path, &old_de) == POSITION_NOT_FOUND)
	goto end_rename;

    pathrelse (&old_entry_path);

    old_inode = old_dentry->d_inode;
    retval = -EPERM;

    if ((old_dir->i_mode & S_ISVTX) && 
	current->fsuid != old_inode->i_uid &&
	current->fsuid != old_dir->i_uid && !fsuser())
	goto end_rename;

    new_inode = new_dentry->d_inode;

    /* look for the new entry in target directory */
    new_de.de_gen_number_bit_string = 0;
    if (reiserfs_find_entry (new_dir, new_dentry->d_name.name, new_dentry->d_name.len, &new_entry_path, &new_de) == POSITION_FOUND) {
	if (!new_inode) {
	    printk ("do_reiserfs_rename: new entry found, inode == 0 though\n");
	}
	/* this entry already exists, we can just set key of object */
	new_entry_added = 1;
    } else {
#ifdef REISERFS_CHECK
	if (new_entry_added) {
	    if (new_de.de_namelen != new_dentry->d_name.len || 
	        memcmp (new_de.de_name, new_dentry->d_name.name, new_de.de_namelen) ||
		de_visible (new_de.de_deh))
		reiserfs_panic (old_dir->i_sb, "vs-7045: reiserfs_rename: suspicious entry found");
	}
#endif /* REISERFS_CHECK */
    }
    pathrelse (&new_entry_path);


    if (new_inode == old_inode) {
	retval = 0;
	goto end_rename;
    }

    if (new_inode && S_ISDIR(new_inode->i_mode)) {
	/* new name exists and points to directory */
	retval = -EISDIR;
	if (!S_ISDIR(old_inode->i_mode))
	    goto end_rename;
	retval = -EINVAL;
	if (is_subdir (new_dentry, old_dentry))
	    goto end_rename;
	retval = -ENOTEMPTY;
	if (!reiserfs_empty_dir (new_inode))
	    goto end_rename;
	retval = -EBUSY;
	if (new_inode->i_count > 1)
	    goto end_rename;
    }

    retval = -EPERM;
    if (new_inode && (new_dir->i_mode & S_ISVTX) && 
	current->fsuid != new_inode->i_uid &&
	current->fsuid != new_dir->i_uid && !fsuser())
	goto end_rename;

    if (S_ISDIR(old_inode->i_mode)) {
	/* old name points to directory */
	retval = -ENOTDIR;
	if (new_inode && !S_ISDIR(new_inode->i_mode))
	    goto end_rename;

	retval = -EINVAL;
	if (is_subdir(new_dentry, old_dentry))
	    goto end_rename;

	retval = -EIO;
	/* directory is renamed, its parent directory will be changed, so find ".." entry */
	dot_dot_de.de_gen_number_bit_string = 0;
	if (reiserfs_find_entry (old_inode, "..", 2, &dot_dot_entry_path, &dot_dot_de) == POSITION_NOT_FOUND)
	    goto end_rename;
	if (dot_dot_de.de_objectid != old_dir->i_ino)
	    goto end_rename;
	pathrelse (&dot_dot_entry_path);

	retval = -EMLINK;
	if (!new_inode && new_dir->i_nlink >= REISERFS_LINK_MAX)
	    goto end_rename;
    }
  
    if (new_entry_added == 0) {
	/* add new entry if we did not do it, but do not mark it as visible */
	retval = reiserfs_add_entry (&th, new_dir, new_dentry->d_name.name, new_dentry->d_name.len, INODE_PKEY (old_inode), &new_de, 0);
	if (retval)
	    goto end_rename;
	//if_in_ram_update_sd (&th, new_dir);
	new_entry_added = 1;
	goto try_again;
    }


    /* look for old name, new name and ".." when renaming directories
     * again. FIXME: no need to do that in most cases.  */
    if (reiserfs_find_entry (old_dir, old_dentry->d_name.name, old_dentry->d_name.len, &old_entry_path, &old_de) == POSITION_NOT_FOUND)
	reiserfs_panic (old_dir->i_sb, "vs-7050: reiserfs_rename: old name not found");
    if (reiserfs_find_entry (new_dir, new_dentry->d_name.name, new_dentry->d_name.len, &new_entry_path, &new_de) == POSITION_NOT_FOUND)
	reiserfs_panic (old_dir->i_sb, "vs-7055: reiserfs_rename: new name not found");
    if (S_ISDIR(old_inode->i_mode)) {
	// FIXME: when we rename directory, ".." has to be changed
	// only when parent directory changes
	if (reiserfs_find_entry (old_inode, "..", 2, &dot_dot_entry_path, &dot_dot_de) == POSITION_NOT_FOUND)
	    reiserfs_panic (old_dir->i_sb, "vs-7060: reiserfs_rename: \"..\" name not found");

	// node containing ".." gets into transaction
	reiserfs_prepare_for_journal(old_inode->i_sb, dot_dot_de.de_bh, 1) ;
    }
    

    /* sanity checking before doing the rename - avoid races */
    if (!entry_points_to_object (new_dentry->d_name.name, new_dentry->d_name.len, &new_de, new_inode))
	goto try_again;
    if (!entry_points_to_object (old_dentry->d_name.name, old_dentry->d_name.len, &old_de, old_inode))
	/* go to re-looking for old entry */
	goto try_again;

    if (S_ISDIR(old_inode->i_mode) && 
       (!entry_points_to_object ("..", 2, &dot_dot_de, old_dir) || 
        !reiserfs_buffer_prepared(dot_dot_de.de_bh))) {
	/* go to re-looking for ".." entry of renamed directory */
	goto try_again;
    } 
    /* ok, all the changes can be done in one fell swoop when we have
       claimed all the buffers needed.*/

    /* make old name hidden */
    //reiserfs_prepare_for_journal(old_inode->i_sb, old_de.de_bh, 1) ;
    // old entry will be deleted in the same transaction, no need to
    // bother about it here
    mark_de_hidden (old_de.de_deh);
    //journal_mark_dirty(&th, old_inode->i_sb, old_de.de_bh) ;

    /* make new name visible and set key of old object (if entry
       existed, it is already visible, if not, key is correct already) */
    // reiserfs_prepare_for_journal(old_inode->i_sb, new_de.de_bh, 1) ;
    // new entry is in transaction any way. No need to do that once
    // more
    mark_de_visible (new_de.de_deh);
    new_de.de_deh->deh_dir_id = INODE_PKEY (old_inode)->k_dir_id;
    new_de.de_deh->deh_objectid = INODE_PKEY (old_inode)->k_objectid;
    //journal_mark_dirty(&th, old_inode->i_sb, new_de.de_bh) ;

    old_dir->i_ctime = old_dir->i_mtime = CURRENT_TIME;
    //if_in_ram_update_sd (&th, old_dir);

    new_dir->i_ctime = new_dir->i_mtime = CURRENT_TIME;
    //if_in_ram_update_sd (&th, new_dir);

    if (new_inode) {
	new_inode->i_nlink--;
	new_inode->i_ctime = CURRENT_TIME;
	//if_in_ram_update_sd (&th, new_inode);
    }
    if (dot_dot_de.de_bh) {
	set_ino_in_dir_entry (&dot_dot_de, INODE_PKEY (new_dir));

	journal_mark_dirty (&th, new_dir->i_sb, dot_dot_de.de_bh);
	old_dir->i_nlink--;
	//if_in_ram_update_sd (&th, old_dir);
	if (new_inode) {
	    new_inode->i_nlink--;
	    //if_in_ram_update_sd (&th, new_inode);
	} else {
	    new_dir->i_nlink++;
	    //if_in_ram_update_sd (&th, new_dir);
	}
    }

    /* ok, renaming done */
    decrement_counters_in_path (&new_entry_path);
    decrement_counters_in_path (&dot_dot_entry_path);

    /* remove old name (it is hidden now) */
    if (reiserfs_cut_from_item (&th, old_dir, &old_entry_path,
				&(old_de.de_entry_key), 0) == -1)
	printk ("reiserfs_rename: could not remove old name\n");
    else {
	old_dir->i_size -= DEH_SIZE + old_de.de_entrylen;
	old_dir->i_blocks = ((old_dir->i_size + 511) >> 9);
	//if_in_ram_update_sd (&th, old_dir);
    }

    reiserfs_update_sd (&th, old_dir, READ_BLOCKS);
    reiserfs_update_sd (&th, new_dir, READ_BLOCKS);
    if (new_inode)
	reiserfs_update_sd (&th, new_inode, READ_BLOCKS);

    /* Update the dcache */
    d_move(old_dentry, new_dentry);
    retval = 0;

 end_rename:
    pathrelse (&old_entry_path);
    pop_journal_writer(windex) ;
    journal_end(&th, old_dir->i_sb, jbegin_count) ;
    return retval;
}
